import os.path;
import zipfile;
import hashlib;
import fileinput;
import glob;

# NOTE: The following drivers cannot be parsed by the current script. They need a manual patch
#  before running this script.

# - namcond1.c uses a define (NODUMPKNOWN) instead of 0. Please change it to 0 before running this
# - cvs.c uses a completely different way of writing ROM codes, that will not work
# - cojag.c needs manual MD5() around the hard-disk checksums

# Import the regular expressions we need
from DriverRegex import rROMSTART, rROMLOAD, rROMLOAD2, rROMEND, rGAME

MISSING_ROM_FNAME = "missing_rom_list.txt";
MISSING_ZIP_FNAME = "missing_zip_list.txt";
PARSE_ERR_FNAME = "parse_err.txt";
mamePath = "mame/";
romsPath = [ ]; 
outputPath = "";
DoIncrementalUpdate = 0;

def Warning(line, w):
    """Emit a warning into a log file, in a standard format."""
    w1 = "%s(%u): WARNING: %s" % (fileinput.filename(), fileinput.filelineno(), w);
    w2 = line[:len(line)-1]; # print without endline

    # Output warning on screen too
    print(w1)
    print(w2)

    mf = file(PARSE_ERR_FNAME, "a");
    mf.write(w1 + "\n");
    mf.write(w2 + "\n");
    mf.close();


def ROMPath(gameName):
    """Search a game ROM archive in all the available ROM paths."""
    for romPath in romsPath:
        name = os.path.join(romPath, gameName + ".zip");
        if os.path.isfile(name):
            return name;
    return "";

def DumpMissingFiles(notFound):
    """Dump missing ROMs into a log file."""
    print("WARNING: %u ROM files not found" % len(notFound))

    mf = file(MISSING_ROM_FNAME, "a"); 
    for g,r,c in notFound:
        mf.write("%-8s:%-12s (%s)\n" % (g,r,c));
    mf.close();        

def DumpMissingZipFile(zipName):
    """Dump a missing ZIP archive into a log file."""
    mf = file(MISSING_ZIP_FNAME, "a");
    mf.write(zipName + "\n");
    mf.close();
        
def ProcessGame(gameName, parentName, romList):
    """Process a single game.

    This function will calculate the SHA-1 checksums of gameName, going through the
    ROM archive. It also takes into account the parent set (supporting both splitted
    and merged sets), and it will return a dictionary associating CRCs to SHA1s.
    """
    
    crcToSha = {};
    zipFileList1 = [];
    zipFileList2 = [];
    parentZipName = "";

    # Search for the archives
    gameZipName = ROMPath(gameName);
    if parentName:
        parentZipName = ROMPath(parentName);

    # If both cannot be found, dump an error and exit
    if not gameZipName and not parentZipName:
        DumpMissingZipFile(gameName);
        return;

    # Open the child & parent ZIP files
    if gameZipName:
        z1 = zipfile.ZipFile(gameZipName);
        zipFileList1 = z1.infolist();

    if parentZipName:
        z2 = zipfile.ZipFile(parentZipName);
        zipFileList2 = z2.infolist();

    # Go through the ROMs we are looking for
    for romName, crc in romList:
        # Search if one of the CRC is found in either the parent or the child
        # ZIP archive. In this way, we identify the file that we need to
        # calculate the SHA1 of.
        l = filter(lambda i: (("%08X"%i.CRC) == crc), zipFileList1);
        if (len(l) != 0):
            zok = z1;
        else:
            l = filter(lambda i: (("%08X"%i.CRC) == crc), zipFileList2);
            if (len(l) != 0):
                zok = z2;
            else:
                zok = None;

        # If the file was found, calculate the SHA1 and insert it
        # into the dictionary
        if zok:         
            curSha1 = hashlib.sha1(zok.read(l[0].filename));
            crcToSha[crc] = curSha1.hexdigest();

    # Close the ZIP archives and exit
    if gameZipName:                
        z1.close();
    if parentZipName:        
        z2.close();

    return crcToSha;        


# Full processing on a driver
def ProcessDriver(driverName):
    """Process a driver, calculating SHA-1s for all the available ROMs.

    This big function is the core of the program. It goes through a driver,
    parsing it to extract CRC informations (remembering file lines where it
    found them for later). It also extracts parent/child informations needed
    to properly handle slip and merged sets. Then, it will calculate the
    SHA-1s for all the ROMS it can find, and it will update the source code
    to insert the new informations.
    """
    
    curGame = "";
    romDB = {};  # empty dictionary
    parentNameDB = {};
    crcToSha = {};
    crcLines = {};
    gameless = [];
  
    fileName = os.path.join(mamePath, driverName);
    outFileName = os.path.join(outputPath, driverName);

    if not os.path.isfile(fileName):
        print("ERROR: File %s not found\n" % fileName);
        return;

    if DoIncrementalUpdate:
        # If the output file already exits, and its date/time is
        #  bigger than the input file, then don't do anything
        if os.path.isfile(outFileName) and os.path.getmtime(fileName) < os.path.getmtime(outFileName):
            print("Skipping driver %s (incremental update)" % driverName);
            return;

    print("Processing driver %s..." % driverName)

    # BUG (in PythonWin IDE)
    fileinput.close();

    # PART 1
    # ------
    # Collect the game+rom+crc information from the driver

    # Go through the file
    for line in fileinput.input(fileName):
        
        # GAME
        m = rGAME.search(line);
        if m:
            game, parent = m.group(1,2);
            if parent != "0":
                parentNameDB[game] = parent;      

        # ROM_START
        m = rROMSTART.search(line);
        if m:
            curGame = m.group(1);
            romDB[curGame] = [];
            continue;

        # ROM_END
        m = rROMEND.search(line);
        if m:
            if not curGame:
                Warning(line, "ROM_END but not ROM_START before");
            curGame = "";
            continue;
                       
        # ROM_LOAD
        m = rROMLOAD.search(line);
        if m:
            romName, romCRC = m.group(1,2);

            # Some driver has extra leading zero(s) before the CRC,
            #  or only 7 numbers for the CRC
            if len(romCRC) > 8:
                romCRC = romCRC.lstrip("0");
            if len(romCRC) < 8:
                romCRC = romCRC.zfill(8);
                
            romCRC = romCRC.upper();
            crcLines[fileinput.filelineno()] = romCRC;

            # CRC=0 means no valid dump known: we just ignore
            if long(romCRC,16) != 0:
                if curGame:
                    romDB[curGame].append((romName, romCRC));
                else:
                    # We did not find a ROM_LOAD before this point. It means that this
                    #  is probably a #define containing some ROM definitions (for BIOS
                    #  for instance). So we add it to a special entry of gameless roms
                    #  and we look for them in all the zipfiles of this driver
                    gameless.append((romName, romCRC));
                        
    # close the file of the driver, we don't need it anymore
    fileinput.close();

    # PART 2
    # ------
    # Compute the SHA1 of all the available files

    # Filter function to remove all the ROMs that were already processed
    def FilterNotFoundRomList(romName, crc):
        return (crc not in crcToSha);
        
    # Compute the SHA1 for all the ROMS (at least, the available ones)
    for gameName, romList in romDB.items():

        # Copy the list of ROMs because we need to modify it
        rl = [x for x in romList];

        # Merge the game-less ROMs to any rom list, so that we look for them
        # in any ZIP of this driver
        rl.extend(gameless);
        
        # Remove from the list of ROMS the ones for which we already calculated a SHA1
        # This is important otherwise all child sets will recompute the parent set ROMS
        rl = filter(FilterNotFoundRomList, rl);
           
        newCrcToSha = ProcessGame(gameName, parentNameDB.get(gameName, ""), rl);

        # Merge the new database with the current            
        if newCrcToSha:
            crcToSha.update(newCrcToSha); 

    # Compute a list of all the ROMS that were not found:
    notFound = [];
    
    nf = filter(FilterNotFoundRomList, gameless);
    for rn,crc in nf:
        notFound.append(("NOGAME", rn, crc));

    for gameName, romList in romDB.items():
        nf = filter(FilterNotFoundRomList, romList);
        for rn,crc in nf:
            notFound.append((gameName, rn, crc));
    notFound.sort();

    # Dump the missing file list
    if len(notFound) > 0:
        DumpMissingFiles(notFound);


    # PART 3
    # ------
    # Using the CRC->SHA1 map, modify the source file
    outFile = file(outFileName, "w");

    # Go through the driver again
    for line in fileinput.input(fileName):

        # If the line does not contain CRC, we do not need to modify it
        if fileinput.filelineno() in crcLines:

            # Since it contains a CRC, must be a ROM_LOAD line. We parse it
            #  using a different regular expression, which returns the whole
            #  part about the CRC in a string
            m = rROMLOAD2.search(line);

            # If there is no match, internal error
            if not m:
                Error(line, "Internal Error: ROMLOAD2 not matching");
            else:
                crc = crcLines[fileinput.filelineno()];
                if long(crc,16) == 0:
                    # If the CRC is zero, it means that there is no valid dump known
                    newPart = "NO_DUMP ";
                else:
                    badrom = "";        
                    if m.group(1).find("BADCRC") != -1:
                        badrom = "BAD_DUMP ";        
                    
                    # Notice that the SHA1 could not exist in the dictionary if the
                    #  ROM was not found. In that case, we just don't output SHA1 informations
                    if crc in crcToSha:                      
                        newPart = "%sCRC(%s) SHA1(%s) " % (badrom, crc.lower(), crcToSha[crc]);
                    else:
                        newPart = "%sCRC(%s) " % (badrom, crc.lower());

                # Merge the new part into the current line (removing the matched part, which
                #  is the old CRC informations)
                line = line[:m.start(1)] + newPart + line[m.end(1):];
            
        # Write the line to the output file    
        outFile.write(line);

    fileinput.close();
    outFile.close();
       

def UpdateDrivers(opts, args):
    """Perform driver updates.

    This is the frontend function, it parses the command line and configure
    the options, and then spawns the real driver update, carried out by
    ProcessDriver.
    """
    
    global mamePath, outputPath;
    
    for o,a in opts:
        if o in ("-h", "--help"):
            usage();
            sys.exit();
        if o in ("-m", "--mame-path"):
            if not os.path.isdir(a):
                print("ERROR: \"%s\" is not a valid directory" % a)
            else:
                mamePath = a;
        if o in ("-r", "--roms-path"):
            if not os.path.isdir(a):
                print("ERROR: \"%s\" is not a valid directory" % a)
            else:
                romsPath.append(a);
        if o in ("-i", "--incremental"):
            global DoIncrementalUpdate;
            DoIncrementalUpdate = 1;

    if len(romsPath) == 0:
        print("ERROR: No ROM path specified!")
        sys.exit(2);

    # Direct path to the drivers
    mamePath = os.path.join(mamePath, "src/drivers");
    outputPath = os.path.join(mamePath, "new_sha1");

    print("MAME path is %s" % mamePath)
    print("ROMS path are %s" % romsPath)
    print("Output path is %s" % outputPath)

    # If the output directory does not exist, create it
    if not os.path.isdir(outputPath):
        try:
            os.mkdir(outputPath);
        except os.error:
            print("Cannot create directory %s" % outputPath)
            sys.exit();

    # Parse the specified drivers
    if len(args) == 1:
        # Get all the C files
        driverList = [os.path.basename(x) for x in glob.glob(os.path.join(mamePath, "*.c"))];
        driverList.sort();
    else:
        driverList = args[1:];

    # Erase the log files if they already exist
    if os.path.isfile(MISSING_ROM_FNAME):
        os.remove(MISSING_ROM_FNAME);
    if os.path.isfile(MISSING_ZIP_FNAME):
        os.remove(MISSING_ZIP_FNAME);
    if os.path.isfile(PARSE_ERR_FNAME):
        os.remove(PARSE_ERR_FNAME);

    for d in driverList:
        ProcessDriver(d);        
